/* $Id: linux.h,v 1.10 1999/08/13 14:34:41 becka Exp $
******************************************************************************

   KGI - Linux specific parts

   Copyright (C) 1997	Steffen Seeger	[seeger@ggi-project.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#ifndef _KGI_SYS_LINUX_H
#define _KGI_SYS_LINUX_H

typedef void * mem_vaddr;      /* the virtual address type     */

#ifdef __KERNEL__
#include <linux/config.h>
#ifdef CONFIG_MODVERSIONS
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#include <linux/version.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <linux/pci.h>
#include <asm/io.h>
#else /* __KERNEL__ */
#include <errno.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#ifndef MAX_NR_CONSOLES
#define MAX_NR_CONSOLES        63
#endif
#endif /* __KERNEL__ */


#ifdef __KERNEL__

#define	IS_CRITICAL	unsigned long __critical_flags 
#define	ENTER_CRITICAL	save_flags(__critical_flags); cli()
#define	LEAVE_CRITICAL	restore_flags(__critical_flags)

#define	LinuxVersionCode(v, p, s) (((v)<<16)+((p)<<8)+(s))

/*	use these macros to allocate memory. 
	kgi_malloc_small() may be used in IRQ environments, but may
	likely fail for big chunks requested there. In general it shall
	be used when allocating small (< PAGESIZE) chunks of memory.
	kgi_malloc/free_huge() are used to allocate big chunks of memory
	such as console backround buffers.
*/

#define kgi_malloc_small(size)	kmalloc(size, GFP_KERNEL)
#define	kgi_free_small(addr)	kfree(addr)

#define	kgi_malloc_huge(size)	vmalloc(size)
#define	kgi_free_huge(addr)	vfree(addr)

#define NOT_TESTED \
	printk(\
    	"\nGood! You hit untested code in %s line %i. Please report whether\n"\
	"there were any difficulties or not. (Simply send this message and a\n"\
	"short explanation of your configuration to " MAINTAINER ".\n\n\n",\
	__FILE__,__LINE__)

/* Defines for struct kgi_mmio_region */
#define	MMIO_RD		VM_READ
#define	MMIO_WR		VM_WRITE 
#define	MMIO_RDWR	(VM_READ | VM_WRITE)
#define	MMIO_SHARED	VM_SHARED


/*	A ..._region is used to describe a region in a physical address space.
**	Per physical address space, there are the following actions possible
**	with a region:
**
**	int ..._check_region(struct ..._region *)
**		Before doing *any* access to a region, you have to check if the
**		region is not served by another driver. check_region() returns
**		EOK (== 0) if noone else claimed (a part) of this region before.
**		or a non-zero value otherwise.
**
**	..._vaddr ..._claim_region(struct ..._region *)
**		This makes a claim that you are going to handle this region
**		from now on. You must have checked a region successfully
**		before claiming it. A non-zero virtual address should be
**		returned.
**
**	..._vaddr ..._free_region(struct ..._region *)
**		Release a claim made earlier on this region. You must not
**		access the region after releasing your claim.
**
**	..._vaddr ..._alloc_region(struct ..._region *r)
**		Allocate and claim a free region of size r->size with r->decode
**		being the address lines decoded. This returns a non-zero
**		virtual address if successful or a zero value otherwise.
**
**	void ..._out<width>(?int<width> val, ..._vaddr vaddr)
**		Do the neccessary actions to write the value 'val' at
**		address 'addr'. 'addr' must be aligned to the <width> of
**		the value being written. The write cycle generated is of width
**		<width>.
**
**	kgi_uint<width> ..._in<width>(..._vaddr vaddr)
**		Do the neccessary actions to reat the cell at
**		address 'addr' which must be alligned to a cell width.
**		The read cycle generated is of width <width>.
*/

/*	Per region, we have a physical and a virtual address space and
**	certain information is stored with each region. As this is
**	common to all regions, let's make life easier.
*/

/*	Intel style in/out I/O space
**
**	The io_in.. / io_out.. functions generate the neccesary hardware
**	actions	to do a read/write cycle in a cards I/O space. (Registers
**	accessed via the in.. / out.. instructions on i386 architecture.)
**
**	The function names encode direction (in == read cycle, out == write
**	cycle) and register width, which is 8, 16 or 32 bits. Any address given
**	__must__ be aligned to register widths, otherwise results may
**	be undefined!
*/

typedef kgiu16 io_addr; 	/* the physical address type	*/
typedef kgiu16 io_vaddr; 	/* the virtual address type	*/

struct io_region 
{
	io_vaddr baseptr;	/* virtual base address		*/
	io_addr base;		/* physical base address	*/
	io_addr size;		/* size of region		*/
	io_addr decode;		/* decoded phys address lines	*/
	char *name;		/* name of the region		*/
};

#define	IO_NULL		((io_vaddr) 0)

#ifdef __sparc__
/* Unimplemented: sparc has MMU managed PIO and it's own function 
   occupy_region() */
#warning Please implement occupy_region() support for Sparc... faking it...
# define	io_check_region(x)	0
#else

static inline int io_check_region(struct io_region *r)
{
	register kgi_sint foo;

	DEBUG3("Checking io_region %s, base = %.4x, size %i.",
	       r->name, r->base, r->size);

	foo = check_region(r->base, r->size);

	if (!foo) {
	  DEBUG2("No holds on io_region %s, base = %.4x, size %i.",
	       r->name, r->base, r->size);
	  return foo;
	}

/* Hopefully we will always be able to unregister the areas used by the 
   in-kernel video code and then re-register them later without altering
   some sort of vaddr needed to access the ports.  Even sparc seems not
   to have a vaddr type thing going on in there.  But we are stuck either
   way because the drivers won't want to deal with a fragmented map of
   their PIO region, unless the underlying drivers can be re-initialized
   to deal with the changed vaddr. */

# ifdef __i386__
	DEBUG1("Trying to free PIO from usual i386 suspects.\n");
	/* These are the regions taken by vgacon and vesafb */
	release_region(0x3c0, 0x20);
	release_region(0x3b0, 0x10);
	release_region(0x3bf, 0x1);
	release_region(0x3d4, 0x2);
	foo = check_region(r->base, r->size);
	if (!foo) {
	  DEBUG1("Success freeing io_region from"
		 "(we hope) kernel vgacon/vesafb driver.\n");
	  return foo;
	}
# else
	DEBUG0("Falsly returning success for io_check_region()\n"
	       "Please define safe PIO areas in kgi/sys/linux.h"
	       "for your architecture!");
	return(0);
# endif /* Architecture switch */

	DEBUG0("Couldn't get PIO ports");
	return(foo);
}

#endif /* Not __sparc__ */



static inline io_vaddr io_claim_region(struct io_region *r)
{
	PASSERT(r->baseptr == IO_NULL, "claim of claimed io_region '%s'!",\
		r->name);

	request_region(r->base, r->size, r->name);
	r->baseptr = r->base;

	DEBUG2("claiming io_region %s, base = %.4x, baseptr = %.4x", \
		r->name, r->base, r->baseptr);

	return r->baseptr;
}

static inline io_vaddr io_free_region(struct io_region *r)
{
	DEBUG2("freeing io_region %s, base = %.4x, baseptr = %.4x", \
		r->name, r->base, r->baseptr);
	release_region(r->base, r->size);
	return (r->baseptr = IO_NULL);
}

static inline io_vaddr io_alloc_region(struct io_region *r)
{
	/*	nothing appropriate under Linux. AFAIK - seeger
	*/
	return IO_NULL;
}

static inline kgiu8 io_in8(io_vaddr addr) 
{
	return inb(addr);
}

static inline kgiu16 io_in16(io_vaddr addr)
{
	return inw(addr);
}

static inline kgiu32 io_in32(io_vaddr addr)
{
	return inl(addr);
}

static inline void io_ins8(io_vaddr addr, void *buf, kgi_uint cnt)
{
	insb(addr, buf, cnt);
}

static inline void io_ins16(io_vaddr addr, void *buf, kgi_uint cnt)
{
	insw(addr, buf, cnt);
}

static inline void io_ins32(io_vaddr addr, void *buf, kgi_uint cnt)
{
	insl(addr, buf, cnt);
}

static inline void io_out8(kgiu8 val, io_vaddr addr)
{
	outb(val, addr);
}

static inline void io_out16(kgiu16 val, io_vaddr addr)
{
	outw(val, addr);
}

static inline void io_out32(kgiu32 val, io_vaddr addr)
{
	outl(val, addr);
}

static inline void io_outs8(io_vaddr addr, void *buf, kgi_uint cnt)
{
	outsb(addr, buf, cnt);
}

static inline void io_outs16(io_vaddr addr, void *buf, kgi_uint cnt)
{
	outsw(addr, buf, cnt);
}

static inline void io_outs32(io_vaddr addr, void *buf, kgi_uint cnt)
{
	outsl(addr, buf, cnt);
}

/*	memory I/O space
**
**	This is the 'normal' shared memory I/O space accessed using the
**	mov instructions on i386 architecture.
*/

typedef kgi_addr mem_addr; 	/* the physical address type	*/

struct mem_region 
{
	mem_vaddr baseptr;	/* virtual base address		*/
	mem_addr base;		/* physical base address	*/
	mem_addr size;		/* size of region		*/
	mem_addr decode;	/* decoded phys address lines	*/
	char *name;		/* name of the region		*/
};

#define	MEM_NULL	((mem_vaddr) 0)

static inline kgi_sint mem_check_region(struct mem_region *r)
{
	/*	no ressource management for physical memory yet. AFAIK - seeger
	*/
	DEBUG2("checking mem_region %s, base %.8x, size %i, is always free!",\
		r->name, r->base, r->size);
	return EOK;
}

static inline mem_vaddr mem_claim_region(struct mem_region *r)
{
	PASSERT(r->baseptr == MEM_NULL, "claim of claimed mem_region '%s'!", \
		r->name);

	if (r->base < 1024 * 1024) {

		r->baseptr = phys_to_virt(r->base);

	} else {

		#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,0)

			r->baseptr = vremap(r->base, r->size);
		#else
			r->baseptr = ioremap(r->base, r->size);
		#endif
	}

	DEBUG2("claiming mem_region %s, base = %.8x, baseptr = %p", \
		r->name, r->base, r->baseptr);
	return r->baseptr;
}

static inline mem_vaddr mem_free_region(struct mem_region *r)
{
	PASSERT(r->baseptr, "free of freed region '%s'!", r->name);

	if (r->base >= 1024 * 1024) {

		#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,0)

			vfree(r->baseptr);
		#else
			iounmap(r->baseptr);
		#endif
	}

	DEBUG2("freeing mem_region %s, base = %.8x, baseptr = %p", \
		r->name, r->base, r->baseptr);
	return (r->baseptr = NULL);
}

static inline mem_vaddr mem_alloc_region(struct mem_region *r)
{
	/*	no ressource management for physical memory yet. AFAIK - seeger
 	*/
	return MEM_NULL;
}

static inline kgiu8 mem_in8(mem_vaddr addr)
{
	return 	readb(addr);
}

static inline kgiu16 mem_in16(mem_vaddr addr)
{
	return readw(addr);
}

static inline kgiu32 mem_in32(mem_vaddr addr)
{
	return readl(addr);
}

static inline void mem_out8(kgiu8 val, mem_vaddr addr)
{
	writeb(val, addr);
}

static inline void mem_out16(kgiu16 val, mem_vaddr addr)
{
	writew(val, addr);
}

static inline void mem_out32(kgiu32 val, mem_vaddr addr)
{
	writel(val, addr);
}

#define	MEM_INS(x) \
static inline void mem_ins##x(mem_vaddr addr, kgiu##x *buf, kgi_uint cnt)	\
{								\
	while (cnt--) {						\
								\
		*(buf++) = mem_in##x(addr);			\
	}							\
}

#define	MEM_OUTS(x) \
static inline void mem_outs##x(mem_vaddr addr, kgiu##x *buf, kgi_uint cnt)	\
{								\
	while (cnt--) {						\
								\
		mem_out##x(*(buf++), addr);			\
	}							\
}

MEM_INS(8)
MEM_INS(16)
MEM_INS(32)

MEM_OUTS(8)
MEM_OUTS(16)
MEM_OUTS(32)

#undef	MEM_OUTS
#undef	MEM_INS

/*	PCI configuration space
**
**	These functions should be used to access the PCI configuration space.
**	As the address is given by the hardware wiring, a registration
**	scheme doesn't make sense. But I/O is pretty useful :)
*/

typedef kgi_addr pcicfg_vaddr;
#define	PCICFG_NULL	((pcicfg_vaddr) -1)
#define	PCICFG_VADDR(bus, slot, fn)	\
	((pcicfg_vaddr)((((bus) & 0xFF) << 24) | (PCI_DEVFN(slot,fn) << 16)))

#define	PCIARGS	(addr >> 24) & 0xFF, (addr >> 16) & 0xFF, addr & 0xFF

#if CONFIG_PCI

static inline kgiu8 pcicfg_in8(pcicfg_vaddr addr)
{
	kgiu8 tmp;
	pcibios_read_config_byte(PCIARGS, &tmp);
	DEBUG2("pcicfg%.8x -> %.2x\n", addr, tmp);
	return tmp;
}

static inline kgiu16 pcicfg_in16(pcicfg_vaddr addr)
{
	kgiu16 tmp;
	pcibios_read_config_word(PCIARGS, &tmp);
	DEBUG2("pcicfg%.8x -> %.4x\n", addr,  tmp);
	return tmp;
}

static inline kgiu32 pcicfg_in32(pcicfg_vaddr addr)
{
	kgiu32 tmp;
	pcibios_read_config_dword(PCIARGS, &tmp);
	DEBUG2("pcicfg%.8x -> %.8x\n", addr, tmp);
	return tmp;
}

static inline void pcicfg_out8(kgiu8 val, pcicfg_vaddr addr)
{
	pcibios_write_config_byte(PCIARGS, val);
	DEBUG2("pcicfg%.8x <- %.2x\n", addr, val);
}

static inline void pcicfg_out16(kgiu16 val, pcicfg_vaddr addr)
{
	pcibios_write_config_word(PCIARGS, val);
	DEBUG2("pcicfg%.8x <- %.4x\n", addr, val);
}

static inline void pcicfg_out32(kgiu32 val, pcicfg_vaddr addr)
{
	pcibios_write_config_dword(PCIARGS, val);
	DEBUG2("pcicfg%.8x <- %.8x\n", addr, val);
}

static inline int pci_get_next_device(pcicfg_vaddr *addr)
{
	kgiu16 vendor,device;

	do {
		(*addr)+=0x10000;	
		/* This steps up function, then device, then bus. */

		if ( 0==((*addr)&0xffff0000) ) 
			return 0;	/* We are through. though luck. */

		vendor=pcicfg_in16(*addr + PCI_VENDOR_ID);
		device=pcicfg_in16(*addr + PCI_DEVICE_ID);

		/* skip over empty slots */
	} while(vendor==0xffff&&device==0xffff);

	return 1;
}

#undef PCIARGS

#else	/* #if CONFIG_PCI */

static inline kgiu8 pcicfg_in8(pcicfg_vaddr addr)
{
	return -1;
}

static inline kgiu16 pcicfg_in16(pcicfg_vaddr addr)
{
	return -1;
}

static inline kgiu32 pcicfg_in32(pcicfg_vaddr addr)
{
	return -1;
}

static inline void pcicfg_out8(kgiu8 val, pcicfg_vaddr addr)
{
}

static inline void pcicfg_out16(kgiu16 val, pcicfg_vaddr addr)
{
}

static inline void pcicfg_out32(kgiu32 val, pcicfg_vaddr addr)
{
}

#endif /* #if CONFIG_PCI */

#endif /* __KERNEL__ */

#endif /* _KGI_SYS_LINUX_H */
